/* calculates value of PI Square
*   Input: None.
*   Output: Value of PI Square & error in actual value,computed value.
*   

*   Author : Vidyut Tarkas
*            vtarkas@mailbox.syr.edu
*/

/***********************************************************

  Please  Run with 1,4 or 9 Nodes.

  **********************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <mpi.h>

#define numOfIntervals 100    /* Number of intervals per process */

double sqrt(double);

void main(int argc, char* argv[])
{
	
	double	PI=3.14159265358979323842643;
	int p;		/*   The Number of processes   */

	int dim_sizes[2];	/*Array that specifies the order
					of each dimension */
	int wrap_around[2];	/*Array that specifies whether
					each dimesion is circular*/
	int reorder = 1;	/*Reorders the processes in the
					new communicator*/
	int coordinates[2];	/*Array that gives the coordinates
					of this process in the grid  */
	int my_grid_rank;	/*The rank of this process in the
					Cartesian Grid*/
	MPI_Comm grid_comm;	
	

	int x1,y1,x2,y2;	/*Left endpoint & right endpoint for this 
					process */
	int gridRowCol;		
	double h, x, y,integral,result;  
	int i,j;

        double f(double x, double y); /* function we are integrating */ 

	/* Initialise MPI */ 
	MPI_Init(&argc, &argv);

	/* Find out how many processes are being used */ 
	MPI_Comm_size(MPI_COMM_WORLD, &p);

	/*The square root of the number of processes = num. of rows and colums*/
	gridRowCol = (int)sqrt(p);

	dim_sizes[0] = gridRowCol;
	dim_sizes[1] = gridRowCol;

	/* Allow wrap-around */
	wrap_around[0] = 1;
	wrap_around[1] = 1;
   
	/* Create a communicator with a Cartesian coordinate structure on the processes. */
	MPI_Cart_create(MPI_COMM_WORLD, 2, dim_sizes, wrap_around,
					reorder, &grid_comm);

	/* Get the rank of this process*/
	MPI_Comm_rank(grid_comm, &my_grid_rank);
 
	/* Return the coordinates of the process with rank my_grid_rank in the Cartesian */
	/*	communicator comm.   */
	MPI_Cart_coords(grid_comm, my_grid_rank, 2, coordinates);
	
	/* Width of an interval */
	h = (double)(1.0/(gridRowCol*numOfIntervals));

	x1 = coordinates[0] * numOfIntervals;
	y1 = coordinates[1] * numOfIntervals;
	x2 = x1 + numOfIntervals;
	y2 = y1 + numOfIntervals;

	for(i = x1; i < x2; i++)
	{ 
		for (j = y1; j < y2; j++)
		{ 
			x = i * h;
			y = j * h;
			integral += (h*h) * (f(x+(0.5*h), y+(0.5*h)));
		}
	}

	/*The root process collects the pi squared values from*/
        /*different processors and sums them up*/
	MPI_Reduce(&integral, &result, 1, MPI_DOUBLE, MPI_SUM,0, grid_comm);

	if(my_grid_rank == 0)
	{
		printf("pi square is : %.16f\n", result);
                printf("Actual pi square is : %.16f\n", PI*PI);
		printf("Error is : %.16f\n", fabs(result - PI*PI));
		printf("resulting pi: %.16f\n", sqrt(result));
	}         
	
        /* shut down MPI */ 
	MPI_Finalize();
}

/* function we are integrating */
double f(double x, double y)
{
	return(16.0/((1.0 + x*x) * (1.0 + y*y)));
}
